"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Client = void 0;
const tslib_1 = require("tslib");
const events_1 = require("events");
const IProvider_1 = require("../interfaces/IProvider");
const buffer_1 = require("buffer");
const Xbqcrypto_1 = require("../utils/Xbqcrypto");
const secp256k1_1 = require("secp256k1");
const bn_js_1 = require("bn.js");
const axios_1 = require("axios");
const retryExecuteFunction_1 = require("../utils/retryExecuteFunction");
const JsonRpcMethods_1 = require("../interfaces/JsonRpcMethods");
class Client extends events_1.EventEmitter {
    constructor(clientConfig, baseAccount) {
        super();
        this.getBlocks = blockIds => { }; // show info about a block (content, finality ...)
        this.getEndorsements = endorsementIds => { }; // show info about a list of endorsements (content, finality ...)
        this.getOperations = operationIds => { }; // show info about a list of operations = (content, finality ...)
        this.buyRolls = (address, rollCount, fee) => { }; // buy rolls with wallet address
        this.sellRolls = (address, rollCount, fee) => { }; // sell rolls with wallet address
        this.sendTransaction = (senderAddress, receiverAddress, amount, fee) => { }; // send coins from a wallet address
        this.sendSmartContract = (senderAddress, bytecode, maxGas, gasPrice, coins, fee) => { }; // create and send an operation containing byte code
        this.readonlySmartContract = (bytecode, maxGas, gasPrice, address) => { }; // execute byte code, address is optionnal. Nothing is really executed on chain
        /* web3/wallet.js module that will under the hood interact with WebExtension, native client or interactively with user */
        this.walletInfo = () => { }; // show wallet info (private keys, public keys, addresses, balances ...)
        this.walletGeneratePrivateKey = () => { }; // generate a private key and add it into the wallet
        this.walletAddPrivateKeys = privateKeys => { }; // add a list of private keys to the wallet
        this.walletRemoveAddresses = addresses => { }; // remove a list of addresses from the wallet
        this.walletSign = (address, string) => { }; // sign provided string with given address (address must be in the wallet)
        this.clientConfig = clientConfig;
        if (this.getPrivateProviders().length === 0) {
            throw new Error("Cannot initialize web3 with no private providers. Need at least one");
        }
        if (this.getPublicProviders().length === 0) {
            throw new Error("Cannot initialize web3 with no public providers. Need at least one");
        }
        if (baseAccount) {
            this.setBaseAccount(baseAccount);
        }
        // bind class methods
        this.getPrivateProviders = this.getPrivateProviders.bind(this);
        this.getProviderForRpcMethod = this.getProviderForRpcMethod.bind(this);
        this.getPublicProviders = this.getPublicProviders.bind(this);
        this.setBaseAccount = this.setBaseAccount.bind(this);
        this.getBaseAccount = this.getBaseAccount.bind(this);
        this.sendJsonRPCRequest = this.sendJsonRPCRequest.bind(this);
        this.executeSC = this.executeSC.bind(this);
        this.signOperation = this.signOperation.bind(this);
        this.computeBytesCompact = this.computeBytesCompact.bind(this);
        // bind api methods
        this.getStatus = this.getPrivateProviders.bind(this);
        this.getAddresses = this.getAddresses.bind(this);
        this.nodeStop = this.nodeStop.bind(this);
    }
    getPrivateProviders() {
        return this.clientConfig.providers.filter((provider) => provider.type === IProvider_1.ProviderType.PRIVATE);
    }
    getPublicProviders() {
        return this.clientConfig.providers.filter((provider) => provider.type === IProvider_1.ProviderType.PUBLIC);
    }
    getProviderForRpcMethod(jsonRpcRequestMethod) {
        switch (jsonRpcRequestMethod) {
            case JsonRpcMethods_1.JSON_RPC_REQUEST_METHOD.GET_ADDRESSES:
            case JsonRpcMethods_1.JSON_RPC_REQUEST_METHOD.GET_STATUS:
            case JsonRpcMethods_1.JSON_RPC_REQUEST_METHOD.SEND_OPERATIONS: {
                return this.getPublicProviders()[0]; //choose the first available public provider
            }
            case JsonRpcMethods_1.JSON_RPC_REQUEST_METHOD.STOP_NODE:
            case JsonRpcMethods_1.JSON_RPC_REQUEST_METHOD.BAN:
            case JsonRpcMethods_1.JSON_RPC_REQUEST_METHOD.UNBAN: {
                return this.getPrivateProviders()[0]; //choose the first available private provider
            }
            default: throw new Error("Unknown Json rpc method");
        }
    }
    setBaseAccount(baseAccount) {
        this.baseAccount = baseAccount;
    }
    getBaseAccount() {
        return this.baseAccount;
    }
    sendJsonRPCRequest(resource, params) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const promise = new Promise((resolve, reject) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                let resp = null;
                const headers = {
                    "Accept": "application/json",
                    'Access-Control-Allow-Origin': '*'
                };
                const body = {
                    "jsonrpc": "2.0",
                    "method": resource,
                    "params": params,
                    "id": 0
                };
                try {
                    resp = yield axios_1.default.post(this.getProviderForRpcMethod(resource).url, body, headers);
                }
                catch (ex) {
                    return resolve({
                        isError: true,
                        result: null,
                        error: new Error('JSON.parse error: ' + String(ex))
                    });
                }
                const responseData = resp.data;
                if (responseData.error) {
                    return resolve({
                        isError: true,
                        result: null,
                        error: new Error(responseData.error)
                    });
                }
                return resolve({
                    isError: false,
                    result: responseData.result,
                    error: null
                });
            }));
            let resp = null;
            try {
                resp = yield promise;
            }
            catch (ex) {
                throw ex;
            }
            // in case of rpc error, rethrow the error
            if (resp.error && resp.error) {
                throw resp.error;
            }
            return resp.result;
        });
    }
    executeSC(contractData, executor) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const signature = this.signOperation(contractData, executor);
            const data = {
                content: {
                    expire_period: contractData.expirePeriod,
                    fee: contractData.fee.toString(),
                    op: {
                        ExecuteSC: {
                            data: Array.from(atob(contractData.contractData), c => c.charCodeAt(0)),
                            max_gas: contractData.maxGas,
                            coins: contractData.coins.toString(),
                            gas_price: contractData.gasPrice.toString()
                        }
                    },
                    sender_public_key: executor.publicKey || this.baseAccount.publicKey
                },
                signature,
            };
            return yield this.sendJsonRPCRequest(JsonRpcMethods_1.JSON_RPC_REQUEST_METHOD.SEND_OPERATIONS, [[data]]);
        });
    }
    signOperation(contractData, executor) {
        // bytes compaction
        const bytesCompact = this.computeBytesCompact(contractData, 3, executor);
        // Hash byte compact
        const hashEncodedData = Xbqcrypto_1.hashSha256(bytesCompact);
        // Signing a digest
        const digest = new bn_js_1.BN(hashEncodedData.valueOf());
        const privateKeyBase58Decoded = Xbqcrypto_1.base58checkDecode(executor.privateKey || this.baseAccount.privateKey);
        const publicKeyBase58Decoded = Xbqcrypto_1.base58checkDecode(executor.publicKey || this.baseAccount.publicKey);
        const base58PrivateKey = new bn_js_1.BN(privateKeyBase58Decoded, 16);
        const base58PublicKey = new bn_js_1.BN(publicKeyBase58Decoded, 16);
        const sig = secp256k1_1.ecdsaSign(digest.toBuffer(), base58PrivateKey.toBuffer());
        const isSigOk = secp256k1_1.ecdsaVerify(sig.signature, digest.toBuffer(), base58PublicKey.toBuffer());
        if (!isSigOk) {
            throw new Error("Malformed signature");
        }
        const r = sig.signature.slice(0, 32);
        const s = sig.signature.slice(32, 64);
        const rr = buffer_1.Buffer.from(Xbqcrypto_1.typedArrayToBuffer(r).toString(), "hex").valueOf();
        const ss = buffer_1.Buffer.from(Xbqcrypto_1.typedArrayToBuffer(s).toString(), "hex").valueOf();
        return Xbqcrypto_1.base58checkEncode(buffer_1.Buffer.concat([rr, ss]));
    }
    computeBytesCompact(contractData, typeId = 3, account) {
        const feeEncoded = buffer_1.Buffer.from(Xbqcrypto_1.varintEncode(contractData.fee));
        const expirePeriodEncoded = buffer_1.Buffer.from(Xbqcrypto_1.varintEncode(contractData.expirePeriod));
        const publicKeyEncoded = Xbqcrypto_1.base58checkDecode(account.publicKey || this.baseAccount.publicKey);
        const typeIdEncoded = buffer_1.Buffer.from(Xbqcrypto_1.varintEncode(typeId));
        const maxGasEncoded = buffer_1.Buffer.from(Xbqcrypto_1.varintEncode(contractData.maxGas));
        const coinsEncoded = buffer_1.Buffer.from(Xbqcrypto_1.varintEncode(contractData.coins));
        const gasPriceEncoded = buffer_1.Buffer.from(Xbqcrypto_1.varintEncode(contractData.gasPrice));
        const dataLengthEncoded = buffer_1.Buffer.from(Xbqcrypto_1.varintEncode(contractData.contractData.length));
        const contractDataEncoded = Uint8Array.from(atob(contractData.contractData), c => c.charCodeAt(0));
        return buffer_1.Buffer.concat([feeEncoded, expirePeriodEncoded, publicKeyEncoded, typeIdEncoded, maxGasEncoded, coinsEncoded, gasPriceEncoded, dataLengthEncoded, contractDataEncoded]);
    }
    // unban a given IP addresses
    unbanIpAddress(ipAddress) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const jsonRpcRequestMethod = JsonRpcMethods_1.JSON_RPC_REQUEST_METHOD.UNBAN;
            if (this.clientConfig.retryStrategyOn) {
                return yield retryExecuteFunction_1.trySafeExecute(this.sendJsonRPCRequest, [jsonRpcRequestMethod, [[ipAddress]]]);
            }
            else {
                return yield this.sendJsonRPCRequest(jsonRpcRequestMethod, [[ipAddress]]);
            }
        });
    }
    // ban a given IP addresses
    banIpAddress(ipAddress) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const jsonRpcRequestMethod = JsonRpcMethods_1.JSON_RPC_REQUEST_METHOD.BAN;
            if (this.clientConfig.retryStrategyOn) {
                return yield retryExecuteFunction_1.trySafeExecute(this.sendJsonRPCRequest, [jsonRpcRequestMethod, [[ipAddress]]]);
            }
            else {
                return yield this.sendJsonRPCRequest(jsonRpcRequestMethod, [[ipAddress]]);
            }
        });
    }
    // stops the node
    nodeStop() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const jsonRpcRequestMethod = JsonRpcMethods_1.JSON_RPC_REQUEST_METHOD.STOP_NODE;
            if (this.clientConfig.retryStrategyOn) {
                return yield retryExecuteFunction_1.trySafeExecute(this.sendJsonRPCRequest, [jsonRpcRequestMethod, []]);
            }
            else {
                return yield this.sendJsonRPCRequest(jsonRpcRequestMethod, []);
            }
        });
    }
    // show staking addresses
    nodeGetStakingAddresses() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const jsonRpcRequestMethod = JsonRpcMethods_1.JSON_RPC_REQUEST_METHOD.GET_STAKING_ADDRESSES;
            if (this.clientConfig.retryStrategyOn) {
                return yield retryExecuteFunction_1.trySafeExecute(this.sendJsonRPCRequest, [jsonRpcRequestMethod, []]);
            }
            else {
                return yield this.sendJsonRPCRequest(jsonRpcRequestMethod, []);
            }
        });
    }
    // remove staking addresses
    nodeRemoveStakingAddresses(addresses) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const jsonRpcRequestMethod = JsonRpcMethods_1.JSON_RPC_REQUEST_METHOD.REMOVE_STAKING_ADDRESSES;
            if (this.clientConfig.retryStrategyOn) {
                return yield retryExecuteFunction_1.trySafeExecute(this.sendJsonRPCRequest, [jsonRpcRequestMethod, [addresses]]);
            }
            else {
                return yield this.sendJsonRPCRequest(jsonRpcRequestMethod, [addresses]);
            }
        });
    }
    // add staking private keys
    nodeAddStakingPrivateKeys(privateKeys) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const jsonRpcRequestMethod = JsonRpcMethods_1.JSON_RPC_REQUEST_METHOD.ADD_STAKING_PRIVATE_KEYS;
            if (this.clientConfig.retryStrategyOn) {
                return yield retryExecuteFunction_1.trySafeExecute(this.sendJsonRPCRequest, [jsonRpcRequestMethod, [privateKeys]]);
            }
            else {
                return yield this.sendJsonRPCRequest(jsonRpcRequestMethod, [privateKeys]);
            }
        });
    }
    /* web3/public.js node instrumentation relying on public Node API */
    // show the status of the node (reachable? number of peers connected, consensus, version, config parameter summary...)
    getStatus() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const jsonRpcRequestMethod = JsonRpcMethods_1.JSON_RPC_REQUEST_METHOD.GET_STATUS;
            if (this.clientConfig.retryStrategyOn) {
                return yield retryExecuteFunction_1.trySafeExecute(this.sendJsonRPCRequest, [jsonRpcRequestMethod, []]);
            }
            else {
                return yield this.sendJsonRPCRequest(jsonRpcRequestMethod, []);
            }
        });
    }
    // get info about a list of addresses (balances, block creation, ...)
    getAddresses(addresses) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const jsonRpcRequestMethod = JsonRpcMethods_1.JSON_RPC_REQUEST_METHOD.GET_ADDRESSES;
            if (this.clientConfig.retryStrategyOn) {
                return yield retryExecuteFunction_1.trySafeExecute(this.sendJsonRPCRequest, [jsonRpcRequestMethod, [[addresses]]]);
            }
            else {
                return yield this.sendJsonRPCRequest(jsonRpcRequestMethod, [[addresses]]);
            }
        });
    }
}
exports.Client = Client;
//# sourceMappingURL=Client.js.map